package paldirvd;

import Jama.Matrix;
import java.util.Vector;

/***/

public class Core {

    private String name="", comment="";
    private int t, g, str, dip;
    private double vol;
    private Vector steps = new Vector();
    private Vector validStep = new Vector();
    private int t_corrected, g_corrected;
    private Vector inclDir, exclDir, inclCirc, exclCirc;
    
    Core(String nameCore, String commentCore, int trend, int plunge, int strike, int diP, double volume){
           name = nameCore.replace("\"","");
           comment = commentCore.replace("\"","");
           t = trend;
           g = plunge;
           str = strike;
           dip = diP;
           vol = volume;
           beddingCorrectionForDrilling();
           inclDir = new Vector(); exclDir = new Vector(); inclCirc = new Vector(); exclCirc = new Vector();
    }

    Step getStepN(int N){
        return (Step)steps.get(N);
    }
    void setStepN(Step stepN){
        steps.add(stepN);
        validStep.add(true);
    }
    Vector getSteps(){
        return steps;
    }
    void setSteps(Vector v){
        steps = v;
        for(int i=0; i<v.size(); i++){
            validStep.add(true);
        }
    }
    void setValidStep(Vector v){
        validStep = v;
    }
    Vector getValidStep(){ // valid positions
        return validStep;
    }

    int getSelectedValidStep(int selectedStep){ // valid steps equivalent position
        int selectedValidStep=0;
        if(!(Boolean)validStep.get(selectedStep))
            return -1;
        for(int i=0; i<selectedStep; i++){
            if((Boolean)validStep.get(i))
                selectedValidStep++;
        }
        return selectedValidStep;
    }
    
    Core getValidCore(){
        Core validCore = new Core(name, comment, t, g, str, dip, vol);
        Vector newSteps = new Vector();
        for(int i=0; i<steps.size(); i++){
            if((Boolean)validStep.get(i)){
                newSteps.add(steps.get(i));
            }
        }
        validCore.setSteps(newSteps);
        validCore.setInclDir(inclDir); validCore.setExclDir(exclDir); validCore.setInclCirc(inclCirc); validCore.setExclCirc(exclCirc);
        return validCore;
    }
    String getName(){ return name; }
    String getComment(){ return comment; }
    int getTrend(){ return t; }
    int getPlunge(){ return g; }
    double getVol(){ return vol; }
    int getStrike(){ return str; }
    int getDip(){ return dip; }
    int getTrendCorrected(){return t_corrected;}
    int getPlungeCorrected(){return g_corrected;}

    void setInclDir (Vector v){ //System.out.println("SET INCL DIR: "+v.size());
        inclDir = v;}
    void setExclDir (Vector v){ //System.out.println("SET EXCL DIR: "+v.size());
        exclDir = v;}
    void setInclCirc (Vector v){ //System.out.println("SET INCL CIRC: "+v.size());
        inclCirc = v;}
    void setExclCirc (Vector v){ //System.out.println("SET EXCL CIRC: "+v.size());
        exclCirc = v;}
    Vector getInclDir() {return inclDir;}
    Vector getExclDir(){return exclDir;}
    Vector getInclCirc(){return inclCirc;}
    Vector getExclCirc(){return exclCirc;}
    
    private Matrix drillingCorrectionMatrix(){
        Matrix A;
        double T = Math.toRadians(t);
        double P = Math.toRadians(g);
        if (T<0 && P<0){
            double[][] array = {{-Math.sin(P), Math.cos(P)*Math.cos(T), Math.cos(P)*Math.sin(T)},
                                {-Math.cos(P), -Math.sin(P)*Math.cos(T), -Math.sin(P)*Math.sin(T)},
                                {0, -Math.sin(T), Math.cos(T)}};
            A = new Matrix(array);
        }else{
            double[][] array = {{Math.sin(P), -Math.cos(P), 0},
                                {Math.cos(P)*Math.cos(T), Math.sin(P)*Math.cos(T), Math.sin(T)},
                                {-Math.cos(P)*Math.sin(T), -Math.sin(P)*Math.sin(T), Math.cos(T)}};
            A = new Matrix(array);
        }
        return A;
    }
    
    private Matrix beddingCorrectionMatrix(){
        double s = Math.toRadians(str);
        double d = Math.toRadians(dip);
        double[][] array = {{1, 0, 0},
                            {0, Math.cos(s), Math.sin(s)},
                            {0, -Math.sin(s), Math.cos(s)}}; 
        Matrix A1 = new Matrix(array); 
        double[][] array2 = {{Math.cos(d), 0, Math.sin(d)},
                             {0, 1, 0},
                             {-Math.sin(d), 0, Math.cos(d)}}; 
        Matrix A2 = new Matrix(array2); 
        double[][] array3 = {{1, 0, 0},
                             {0, Math.cos(s), -Math.sin(s)},
                             {0, Math.sin(s), Math.cos(s)}}; 
        Matrix A3 = new Matrix(array3); 
        Matrix A = A1.times(A2).times(A3);
        return A;
    }
    
  
    public Core drillingCorrection(){
        Core coreDC =  new Core(name, comment, t, g, str, dip, vol);
        Matrix A = drillingCorrectionMatrix();
        double[][] array = new double[steps.size()][3];
        Step stepN; Position pos; 
        
        for(int i=0; i<steps.size(); i++){
            array[i][0] = ((Step)steps.get(i)).getPos().getX();
            array[i][1] = ((Step)steps.get(i)).getPos().getY();
            array[i][2] = ((Step)steps.get(i)).getPos().getZ();
        }
        Matrix P = new Matrix(array);
        P = A.times(P.transpose());
 
        for(int i=0; i<steps.size(); i++){
            pos = new Position (P.get(0,i), P.get(1,i), P.get(2,i), ((Step)steps.get(i)).getPos().getError());
            stepN = new Step(((Step)steps.get(i)).getTemp(), pos);
            coreDC.setStepN(stepN);
        }
        coreDC.setValidStep(validStep);
        return coreDC;
    }
    
    public Core beddingCorrection(){

        Core coreBC = new Core(name, comment, t, g, str, dip, vol);

        Matrix A = beddingCorrectionMatrix();
        double[][] array = new double[steps.size()][3];
        double[][] array2 = new double[3][1]; Matrix P2;
        Step stepN; Position pos, pos2, posMean, posMean2; VD vd, vd2;
        Vector incl_dir=new Vector(), excl_dir=new Vector(), incl_circ=new Vector(), excl_circ=new Vector();
        
        for(int i=0; i<steps.size(); i++){
            array[i][0] = ((Step)steps.get(i)).getPos().getX();
            array[i][1] = ((Step)steps.get(i)).getPos().getY();
            array[i][2] = ((Step)steps.get(i)).getPos().getZ();
        }
        Matrix P = new Matrix(array);
        P = A.times(P.transpose());
        for(int i=0; i<steps.size(); i++){
            pos = new Position (P.get(0,i), P.get(1,i), P.get(2,i), ((Step)steps.get(i)).getPos().getError());
            stepN = new Step(((Step)steps.get(i)).getTemp(), pos);
            coreBC.setStepN(stepN);
        }
        coreBC.setValidStep(validStep);

        //PCA or VD
        for(int i=0; i<inclDir.size(); i++){
            vd=(VD)inclDir.get(i);
            pos=vd.getPos();
            array2[0][0]=pos.getX(); array2[1][0]=pos.getY(); array2[2][0]=pos.getZ();
            P2 = new Matrix(array2);
            P2 = A.times(P2);
            pos2 = new Position (P2.get(0,0), P2.get(1,0), P2.get(2,0), pos.getError());
            vd2 = new VD(vd.getNameCore(), vd.getResDif(), vd.getPlaneLine(), vd.getStepIni(), vd.getStepEnd(), vd.getNumSteps(), pos2, pos2);
            incl_dir.add(vd2);
        }
        for(int i=0; i<exclDir.size(); i++){
            vd=(VD)exclDir.get(i);
            pos=vd.getPos();
            array2[0][0]=pos.getX(); array2[1][0]=pos.getY(); array2[2][0]=pos.getZ();
            P2 = new Matrix(array2);
            P2 = A.times(P2);
            pos2 = new Position (P2.get(0,0), P2.get(1,0), P2.get(2,0), pos.getError());
            
            posMean=vd.getPosMean();
            array2[0][0]=posMean.getX(); array2[1][0]=posMean.getY(); array2[2][0]=posMean.getZ();
            P2 = new Matrix(array2);
            P2 = A.times(P2);
            posMean2 = new Position (P2.get(0,0), P2.get(1,0), P2.get(2,0), pos.getError());
            
            vd2 = new VD(vd.getNameCore(), vd.getResDif(), vd.getPlaneLine(), vd.getStepIni(), vd.getStepEnd(), vd.getNumSteps(), pos2, posMean2);
            excl_dir.add(vd2);
        }
        for(int i=0; i<inclCirc.size(); i++){
            vd=(VD)inclCirc.get(i);
            pos=vd.getPos();
            array2[0][0]=pos.getX(); array2[1][0]=pos.getY(); array2[2][0]=pos.getZ();
            P2 = new Matrix(array2);
            P2 = A.times(P2);
            pos2 = new Position (P2.get(0,0), P2.get(1,0), P2.get(2,0), pos.getError());
            vd2 = new VD(vd.getNameCore(), vd.getResDif(), vd.getPlaneLine(), vd.getStepIni(), vd.getStepEnd(), vd.getNumSteps(), pos2, pos2);
            incl_circ.add(vd2);
        }
        for(int i=0; i<exclCirc.size(); i++){
            vd=(VD)exclCirc.get(i);
            pos=vd.getPos();
            array2[0][0]=pos.getX(); array2[1][0]=pos.getY(); array2[2][0]=pos.getZ();
            P2 = new Matrix(array2);
            P2 = A.times(P2);
            pos2 = new Position (P2.get(0,0), P2.get(1,0), P2.get(2,0), pos.getError());
            vd2 = new VD(vd.getNameCore(), vd.getResDif(), vd.getPlaneLine(), vd.getStepIni(), vd.getStepEnd(), vd.getNumSteps(), pos2, pos2);
            excl_circ.add(vd2);
        }
        coreBC.setInclDir(incl_dir); coreBC.setExclDir(excl_dir); coreBC.setInclCirc(incl_circ); coreBC.setExclCirc(excl_circ);

        return coreBC;
    }
    
    private void beddingCorrectionForDrilling(){
        double[][] array = new double[1][3];
        Matrix A = beddingCorrectionMatrix();
        Position drillingPos = new Position(t, g, 0);
        array[0][0] = drillingPos.getX();
        array[0][1] = drillingPos.getY();
        array[0][2] = drillingPos.getZ();
        Matrix P = new Matrix(array);
        P = A.times(P.transpose());
        drillingPos = new Position(P.get(0,0), P.get(1,0), P.get(2,0),0);
        t_corrected = (int)drillingPos.getDec();
        g_corrected = (int)drillingPos.getInc();
    }
      
}
