
package paldirvd;

import Jama.EigenvalueDecomposition;
import Jama.Matrix;
import java.util.Arrays;
import java.util.HashMap;
import java.util.Vector;

public class Statistics {
    
    
    
    private static double A = 0.05;
    private static String ALL = "All";
    private static String POS = "Positive";
    private static String NEG = "Negative";
    private static String ROT = "Pos + Neg Rotated";
            
    public static Core stacking(Vector site){
        
        Vector stacking = new Vector();
        Vector steps; 
        Step step;
        HashMap stepStacking = new HashMap() {};
        PosStacking posStacking;
        int strike, dip;

        strike = ((Core)site.get(0)).getStrike(); dip = ((Core)site.get(0)).getDip();
        for(int i=0; i<site.size(); i++){
            steps = ((Core)site.get(i)).getSteps(); //No se están considerando vectores unitarios
            for(int j=0; j<steps.size(); j++){
                step = (Step)steps.get(j);
                if(stepStacking.containsKey(step.getTemp())){
                    posStacking = (PosStacking) stepStacking.get(step.getTemp());
                    posStacking.setPos(step.getPos());
                }else{
                    posStacking = new PosStacking(step.getPos());
                }
                stepStacking.put(step.getTemp(),posStacking);
            }
        }
        Object[] key = stepStacking.keySet().toArray();
        Arrays.sort(key);
        for(int i=0; i<stepStacking.size(); i++){
            stepStacking.get(key[i]);
            step = new Step(((Double)key[i]).doubleValue(), ((PosStacking)stepStacking.get(key[i])).getMeanPos(), ((PosStacking)stepStacking.get(key[i])).getN());
            stacking.add(step);
        }
        
        Core core = new Core("STACKING", "",0, 0, strike, dip, 0);
        core.setSteps(stacking);
        return core;
    }
    
 
    
    public static Vector fisherMean(Vector v){
        
        Vector fisher = new Vector();
        double x=0, y=0, z=0, error=0, x1=0, y1=0, z1=0, error1=0, x2=0, y2=0, z2=0, error2=0, x3=0, y3=0, z3=0, error3=0;
        int n1=0, n2=0;
        double k;

        for(int i=0; i<v.size(); i++){
            Position pos = ((Position)v.get(i));
            //pos=pos.normalize(); //It's not necessary, it's already normalized

            x = x + pos.getX();
            y = y + pos.getY();
            z = z + pos.getZ();
            error = error + pos.getError();
            if(pos.getInc() >= 0 ){
                x1 = x1 + pos.getX();
                y1 = y1 + pos.getY();
                z1 = z1 + pos.getZ();
                error1 = error1 + pos.getError();
                n1++;
            }else{
                x2 = x2 + pos.getX();
                y2 = y2 + pos.getY();
                z2 = z2 + pos.getZ();
                error2 = error2 + pos.getError();
                n2++;
                pos = pos.inverse();
            }

            x3 = x3 + pos.getX();
            y3 = y3 + pos.getY();
            z3 = z3 + pos.getZ();
            error3 = error3 + pos.getError();
        }
        Position pos = new Position(x,y,z,error); // positive and negative 
        fisher.add(pos.normalize());
        
        Position pos1 = new Position(x1,y1,z1,error1); // positive
        fisher.add(pos1.normalize());
        
        Position pos2 = new Position(x2,y2,z2,error2); // negative
        fisher.add(pos2.normalize());
        
        Position pos3 = new Position(x3,y3,z3,error3); // positve and negative rotated
        fisher.add(pos3.normalize());
        
        //System.out.println("Fisher All");
        if((n1+n2)>0){
        k = kFisher(pos, n1+n2, v, ALL);
        fisher.add(k);
        fisher.add(a95Fisher(pos, n1+n2, k));
        }else{fisher.add(0); fisher.add(0);}
        
        //System.out.println("Fisher Positive");
        if(n1>0){
        k = kFisher(pos1, n1, v, POS);
        fisher.add(k);
        fisher.add(a95Fisher(pos1, n1, k));
        fisher.add(n1);
        }else{fisher.add(0); fisher.add(0); fisher.add(0);}

        //System.out.println("Fisher Negative");
        if(n2>0){
        k = kFisher(pos2, n2, v, NEG);
        fisher.add(k);
        fisher.add(a95Fisher(pos2, n2, k));
        fisher.add(n2);
        }else{fisher.add(0); fisher.add(0); fisher.add(0);}

        //System.out.println("Fisher Negative and Positive Rotated");
        if((n1+n2)>0){
        k = kFisher(pos3, n1+n2, v, ROT);
        fisher.add(k);
        fisher.add(a95Fisher(pos3, n1+n2, k));
        fisher.add(n1+n2);
        }else{fisher.add(0); fisher.add(0); fisher.add(0);}
               
        return fisher;
    }
    
    private static double kFisher(Position pos_mean, int n, Vector v, String tipo){
        boolean goAhead=true;
        double Xi[] = new double[(int)n];
        double x_mean = pos_mean.getX(), y_mean = pos_mean.getY(), z_mean = pos_mean.getZ();
        double xi = 0, yi = 0, zi = 0, inti=0; Position posi;
        double R = pos_mean.getInt();
        double k = 0, k_mle = n/(n-R), ci;
        
        //System.out.println("Fisher Dec: " + pos_mean.getDec() + ", Inc: " + pos_mean.getInc());
       //System.out.println("n: " + n + ", R: " + R + ", R/n: " + R/n + ", (1-1/n)*kmle: " + ((1-1/n)*k_mle) + ", x: " + x_mean + ", y: " + y_mean + ", z: " + z_mean);


        //CREO QUE STEREONET CALCULA LA K BUENA Y A95 SOLO SI K>3
        if(R/n >= 0.95){
            if(n >= 16){
                k = (1-1/n)*k_mle; // => k = (n-1)/(n-R);
                //System.out.println("k R/n>=0.95, n>=16: " + k);
            }else{
                k = Math.pow(1-1/n,2)*k_mle;
                //System.out.println("k R/n>=0.95, n<16: " + k);
            }
        } else{
            int j=0;
            for(int i=0; i<v.size(); i++){
                posi = ((Position)v.get(i)) ;
                goAhead = true;
                if(posi.getInc()<0){
                    if(ROT.equals(tipo)){
                        posi = posi.inverse();
                    }
                    if(POS.equals(tipo))
                        goAhead = false;                    
                }else{
                    if(NEG.equals(tipo))
                        goAhead = false;
                }
                if(goAhead){
                    xi =  posi.getX() ;
                    yi =  posi.getY();
                    zi =  posi.getZ();
                    //inti = posi.getInt();
                    inti=1;
                    Xi[j] = 1 - (x_mean*xi+y_mean*yi+z_mean*zi)/(R*inti); 
                    //Xi=1-cos(o) (p.130 Statistical analysis of spherical data)
                    j++;
                }
            }
            Arrays.sort(Xi);
            k=0; double ii, N=n; 
            for(int i=0; i<n; i++){
                ii = i+1;
                ci = (1/N-2*ii/(N*N*(N+1)));
                k = k + ci * Xi[i]; 
            }
            k = 1/k;
            //System.out.println("k R/n<0.95: " + k);
        }
        k = Math.round(k*10)/10d;
        //System.out.println("n: " + n + ", R: " + R + ", R/n: " + R/n + ", kmle: "+k_mle + ", k: "+k);
        return k;   
    }
    private static double a95Fisher(Position pos, double n, double k){
        double a95 = 90;
        double R = pos.getInt();
        //System.out.println("n: " +n+", R: "+R+", k: "+k);
        /*if(k>5){
            a95 = Math.toDegrees(Math.acos(1-((n-R)/R)*(Math.pow(1/A, 1/(n-1))-1)));
            //System.out.println("A95 k>5: " + a95);System.out.println("Por partes 1: " + (Math.pow(1/A, 1/(n-1))-1)); System.out.println("Por partes 2: " + (1-((n-R)/R)*(Math.pow(1/A, 1/(n-1))-1)));
        }else if(n>=30){
            a95 = Math.toDegrees(Math.acos(1+Math.log(A)/(k*R)));
            //System.out.println("A95 k<5, n>=30: " + a95);
        }else if(n>8){ //¿?
            double Rz = pos.getZ();
            double R1 = Math.pow(Rz*Rz-2/3*n*Math.log(A),1/2);
            double R2 = n*(1-Math.pow(A,1/(n-1))) + Rz*Math.pow(A,1/(n-1));
            if(Rz<n/4)
                a95 = n * R1;
            else if(Rz<=3*n/5)
                a95 = n * (R1+R2)/2;
            else
                a95 = n * R2;
            //System.out.println("A95 k<5, 8<n<30: " + a95);
        } */
        if(Math.abs(1-((n-R)/R)*(Math.pow(1/A, 1/(n-1))-1))<1)
        a95 = Math.toDegrees(Math.acos(1-((n-R)/R)*(Math.pow(1/A, 1/(n-1))-1))); //first aproximation for all cases
        a95 = Math.round(a95*10)/10d;
        if(a95>90) a95=90;
        return a95;
    }
    
    public static Vector bingham(Vector v){ // Bingham statistics (equivalent to Principal Components Analysis (PCA))
        
        Vector bingham = new Vector();        
        double x=0, y=0, z=0, XX=0, XY=0, XZ=0, YY=0, YZ=0, ZZ=0;
        double a_var=0, b_var=0, a_mean=0, b_mean=0, ga, gb, s1, s2;
        int n = v.size();
        
       
        for(int i=0; i<n; i++){
            Position pos = ((Position)v.get(i));
            //pos=pos.normalize();
            x = pos.getX();
            y = pos.getY();
            z = pos.getZ();
            XX = XX + x*x;
            XY = XY + x*y;
            XZ = XZ + x*z;
            YY = YY + y*y;
            YZ = YZ + y*z;
            ZZ = ZZ + z*z;
        }
        double[][] array_p = {{XX, XY, XZ},
                               {XY, YY, YZ},
                               {XZ, YZ, ZZ}};
        Matrix P = new Matrix(array_p);
        
        EigenvalueDecomposition eigen = P.eig(); 
        
        double e3 = eigen.getD().get(0,0); double e2 = eigen.getD().get(1,1); double e1 = eigen.getD().get(2,2);
        double suma_e = e1+e2+e3;
        e1 = e1/suma_e; e2 = e2/suma_e; e3 = e3/suma_e;
        e1 = Math.round(e1*1000)/1000d; e2 = Math.round(e2*1000)/1000d; e3 = Math.round(e3*1000)/1000d;
        bingham.add(e3); bingham.add(e2); bingham.add(e1); 
        //System.out.println("Bingham, e3: " + e3 + ", e2: " + e2 + ", e1: " + e1 );
        
        for(int i=0; i<3; i++){
            x = eigen.getV().get(0,i);
            y = eigen.getV().get(1,i);
            z = eigen.getV().get(2,i);
            Position pos = new Position(x,y,z,0);
            if(pos.getInc()<0){
                  pos = pos.inverse();
            }
            bingham.add(pos);
       }
        
        //Elliptical confidence cone for v3
        a_mean=0; b_mean=0; a_var=0; b_var=0;
        for(int i=0; i<n; i++){
            Position pos = ((Position)v.get(i));
            x = pos.getX() - eigen.getV().get(0,0);
            y = pos.getY() - eigen.getV().get(1,0);
            z = pos.getZ() - eigen.getV().get(2,0);

            a_mean = a_mean + y;
            b_mean = b_mean + z;
            a_var = a_var + y*y;
            b_var = b_var + z*z;
        }a_mean = a_mean/n; b_mean = b_mean/n; a_var = a_var/n; b_var = b_var/n;
          
        ga = -2*Math.log(A)/(n*a_mean*a_mean);
        gb = -2*Math.log(A)/(n*a_mean*b_mean);
                      
        s1 = Math.toDegrees(Math.asin(Math.sqrt(a_var)/ga));
        s2 = Math.toDegrees(Math.asin(Math.sqrt(b_var)/gb));
        //System.out.println("Confidence cone v3, s1: " + s1 + ", s2: " + s2 );
        s1 = Math.round(s1*10)/10d; s2 = Math.round(s2*10)/10d;
        bingham.add(s1); bingham.add(s2);
        
        //Elliptical confidence cone for v1 (p.138 Statistical analysis of spherical data)
        for(int i=0; i<n; i++){
            Position pos = ((Position)v.get(i));
            x = pos.getX() - eigen.getV().get(0,2);
            y = pos.getY() - eigen.getV().get(1,2);
            z = pos.getZ() - eigen.getV().get(2,2);

            a_mean = a_mean + y;
            b_mean = b_mean + z;
            a_var = a_var + y*y;
            b_var = b_var + z*z;
        }a_mean = a_mean/n; b_mean = b_mean/n; a_var = a_var/n; b_var = b_var/n;
          
        ga = -2*Math.log(A)/(n*a_mean*a_mean);
        gb = -2*Math.log(A)/(n*a_mean*b_mean);
                      
        s1 = Math.toDegrees(Math.asin(Math.sqrt(a_var)/ga));
        s2 = Math.toDegrees(Math.asin(Math.sqrt(b_var)/gb));
        //System.out.println("Confidence cone v1, s1: " + s1 + ", s2: " + s2 );
        s1 = Math.round(s1*10)/10d; s2 = Math.round(s2*10)/10d;
        bingham.add(s1); bingham.add(s2);
        
        return bingham;
    }

    
    
    private static class PosStacking {
        
        private int i;
        private Position position, posMean;
        private Vector v, vNorm;
        
        PosStacking(Position pos){
            position = pos;
            posMean = pos.normalize();
            i = 1;
            v = new Vector(); v.add(pos);
            vNorm = new Vector(); vNorm.add(pos.normalize());
        }
        Position getPos(){return position;}
        int getN(){return i;}
        void setPos(Position pos){ 
            sumPos(pos);
            v.add(pos);
            vNorm.add(pos.normalize());
        }
        void sumPos(Position pos){
            double x,y,z,error;
            x = position.getX() + pos.getX();
            y = position.getY() + pos.getY();
            z = position.getZ() + pos.getZ();
            error = position.getError() + pos.getError();
            position = new Position(x,y,z,error);
            pos=pos.normalize();
            x = posMean.getX() + pos.getX();
            y = posMean.getY() + pos.getY();
            z = posMean.getZ() + pos.getZ();
            posMean = new Position(x,y,z,error);
            i++;
        }
        Position getMeanPos(){
            double x,y,z,error;
            x = position.getX() / i;
            y = position.getY() / i;
            z = position.getZ() / i;
            position = new Position(x,y,z,0);
            //error = position.getError() / i;
            //STACKING ERROR HAS NOTHING TO DO WITH MEASUREMENT ERRORS. STACKING ERROR IS THE ALFA 95 OF ALL POSITIONS.
            //FISHER MEAN IS FOR UNIT VECTORS
            double k = kFisher(posMean,i,vNorm,ALL);
            double a95 = a95Fisher(posMean,i,k);
            //System.out.println("N: "+i+", k: "+k+", a95: "+a95);
            error=a95;
            //OPTION 2
            error=getErrorB();
            //System.out.println("a95: "+a95+", errorB: "+getErrorB());
            position = new Position(x,y,z,error);
            return position;
        }
        
        double getErrorB(){ //B ERROR, USED IN GAMMSTACK
            //takes into account the intensity of vectors
            double B, E, I, x=0, y=0, z=0;
            for(int j=0; j<v.size(); j++){
                x = x+Math.abs(((Position)v.get(j)).getX()-position.getX());
                y = y+Math.abs(((Position)v.get(j)).getY()-position.getY());
                z = z+Math.abs(((Position)v.get(j)).getZ()-position.getZ());
                //System.out.println("X: "+((Position)v.get(j)).getX()+", Xmean: "+position.getX());
            }
            x=x/i; y=y/i; z=z/i;
            E = Math.sqrt(x*x+y*y+z*z);
            I = position.getInt();
            B = Math.toDegrees(Math.atan(E/I));
            //System.out.println("B: "+B+", E: "+E+", I: "+I);
            return B; 
        }
    }
    

    
}

